package de.fhhannover.inform.trust.ifmapj.binding.dom;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import util.Pair;
import de.fhhannover.inform.trust.ifmapj.binding.IfmapStrings;
import de.fhhannover.inform.trust.ifmapj.binding.Marshaller;
import de.fhhannover.inform.trust.ifmapj.exception.MarshalException;
import de.fhhannover.inform.trust.ifmapj.identifier.AccessRequest;
import de.fhhannover.inform.trust.ifmapj.identifier.Device;
import de.fhhannover.inform.trust.ifmapj.identifier.Identifier;
import de.fhhannover.inform.trust.ifmapj.identifier.IdentifierWithAd;
import de.fhhannover.inform.trust.ifmapj.identifier.Identity;
import de.fhhannover.inform.trust.ifmapj.identifier.IdentityType;
import de.fhhannover.inform.trust.ifmapj.identifier.IpAddress;
import de.fhhannover.inform.trust.ifmapj.identifier.IpAddressType;
import de.fhhannover.inform.trust.ifmapj.identifier.MacAddress;
import de.fhhannover.inform.trust.ifmapj.log.IfmapJLog;
import de.fhhannover.inform.trust.ifmapj.messages.EndSessionRequest;
import de.fhhannover.inform.trust.ifmapj.messages.IdentifierHolder;
import de.fhhannover.inform.trust.ifmapj.messages.MetadataHolder;
import de.fhhannover.inform.trust.ifmapj.messages.MetadataLifetime;
import de.fhhannover.inform.trust.ifmapj.messages.NamespaceDeclarationHolder;
import de.fhhannover.inform.trust.ifmapj.messages.NewSessionRequest;
import de.fhhannover.inform.trust.ifmapj.messages.PollRequest;
import de.fhhannover.inform.trust.ifmapj.messages.PublishDelete;
import de.fhhannover.inform.trust.ifmapj.messages.PublishElement;
import de.fhhannover.inform.trust.ifmapj.messages.PublishNotify;
import de.fhhannover.inform.trust.ifmapj.messages.PublishRequest;
import de.fhhannover.inform.trust.ifmapj.messages.PublishUpdate;
import de.fhhannover.inform.trust.ifmapj.messages.PurgePublisherRequest;
import de.fhhannover.inform.trust.ifmapj.messages.RenewSessionRequest;
import de.fhhannover.inform.trust.ifmapj.messages.Request;
import de.fhhannover.inform.trust.ifmapj.messages.SearchHolder;
import de.fhhannover.inform.trust.ifmapj.messages.SearchRequest;
import de.fhhannover.inform.trust.ifmapj.messages.SubscribeDelete;
import de.fhhannover.inform.trust.ifmapj.messages.SubscribeElement;
import de.fhhannover.inform.trust.ifmapj.messages.SubscribeRequest;
import de.fhhannover.inform.trust.ifmapj.messages.SubscribeUpdate;

/**
 * A simple Marshaller using JAVA's DOM.
 * 
 * First a document object tree is created from the given {@link XmlMarshalable}
 * object. Second this document object tree is marshalled into an {@link InputStream}.
 * 
 * @author aw
 *
 */
public class DomMarshaller implements Marshaller {
	
	private static DocumentBuilderFactory sDocumentBuilderFactory;
	private static TransformerFactory sTransformerFactory;

	private DocumentBuilder mDocumentbuilder;
	private Transformer trans;
	
	static {
		sTransformerFactory = TransformerFactory.newInstance();
		sDocumentBuilderFactory = DocumentBuilderFactory.newInstance();
		sDocumentBuilderFactory.setNamespaceAware(true);
	}
	
	public DomMarshaller() {
		try {
			mDocumentbuilder = sDocumentBuilderFactory.newDocumentBuilder();
			trans = sTransformerFactory.newTransformer();
		} catch (ParserConfigurationException e) {
			IfmapJLog.error("Could not get a new document builder ["
					+ e.getMessage() + "]");
			throw new RuntimeException(e);
		} catch (TransformerConfigurationException e) {
			IfmapJLog.error("Could not get a new Transformer instance ["
					+ e.getMessage() + "]");
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public InputStream marshalNewSessionRequest(NewSessionRequest nr)
			throws MarshalException {
		Integer mprs = nr.getMaxPollResultSize();
		XmlMarshalable xmlEnv = createSoapEnvelope();
		XmlMarshalable xmlBody = attachSoapBody(xmlEnv);
		XmlMarshalable xmlNr = createIfmapNamespaceMarshalable(
				IfmapStrings.NEW_SESSION_REQ_EL_NAME);
		
		if (mprs != null)
			xmlNr.addXmlAttribute(IfmapStrings.NEW_SESSION_REQ_ATTR_MPRS,
					mprs.toString());
				
		xmlBody.addXmlChild(xmlNr);
		return marshal(xmlEnv);
	}


	@Override
	public InputStream marshalEndSessionRequest(EndSessionRequest er)
			throws MarshalException {
		XmlMarshalable xmlEnv = createSoapEnvelope();
		XmlMarshalable xmlBody = attachSoapBody(xmlEnv);
		XmlMarshalable xmlEr = createIfmapNamespaceMarshalable(
				IfmapStrings.END_SESSION_REQ_EL_NAME);
		
		xmlBody.addXmlChild(xmlEr);
		addSessionId(xmlEr, er);
		return marshal(xmlEnv);
	}

	@Override
	public InputStream marshalPurgePublisherRequest(PurgePublisherRequest pr)
			throws MarshalException {
		XmlMarshalable xmlEnv = createSoapEnvelope();
		XmlMarshalable xmlBody = attachSoapBody(xmlEnv);
		XmlMarshalable xmlPp = createIfmapNamespaceMarshalable(
				IfmapStrings.PURGE_PUBLISHER_REQ_EL_NAME);
		xmlPp.addXmlAttribute(IfmapStrings.PUBLISHER_ID_ATTR, pr.getPublisherId());
		xmlBody.addXmlChild(xmlPp);
		addSessionId(xmlPp, pr);
		return marshal(xmlEnv);
	}

	@Override
	public InputStream marshalRenewSessionRequest(RenewSessionRequest rr)
			throws MarshalException {
		XmlMarshalable xmlEnv = createSoapEnvelope();
		XmlMarshalable xmlBody = attachSoapBody(xmlEnv);
		XmlMarshalable xmlRr = createIfmapNamespaceMarshalable(
				IfmapStrings.RENEW_SESSION_REQ_EL_NAME);
		
		xmlBody.addXmlChild(xmlRr);
		addSessionId(xmlRr, rr);
		return marshal(xmlEnv);
	}

	@Override
	public InputStream marshalPollRequest(PollRequest pr)
			throws MarshalException {
		XmlMarshalable xmlEnv = createSoapEnvelope();
		XmlMarshalable xmlBody = attachSoapBody(xmlEnv);
		XmlMarshalable xmlPr = createIfmapNamespaceMarshalable(
				IfmapStrings.POLL_REQ_EL_NAME);
		
		xmlBody.addXmlChild(xmlPr);
		addSessionId(xmlPr, pr);
		return marshal(xmlEnv);
	}


	@Override
	public InputStream marshalPublishRequest(PublishRequest pr)
			throws MarshalException {
		XmlMarshalable xmlEnv = createSoapEnvelope();
		XmlMarshalable xmlBody = attachSoapBody(xmlEnv);
		XmlMarshalable xmlPr = createIfmapNamespaceMarshalable(
				IfmapStrings.PUBLISH_REQ_EL_NAME);
		
		transformPublishRequest(xmlPr, pr);
		
		xmlBody.addXmlChild(xmlPr);
		addSessionId(xmlPr, pr);
		return marshal(xmlEnv);
	}



	@Override
	public InputStream marshalSubscribeRequest(SubscribeRequest sr)
			throws MarshalException {
		XmlMarshalable xmlEnv = createSoapEnvelope();
		XmlMarshalable xmlBody = attachSoapBody(xmlEnv);
		XmlMarshalable xmlSr = createIfmapNamespaceMarshalable(
				IfmapStrings.SUBSCRIBE_REQ_EL_NAME);
		
		transformSubscribeRequest(xmlSr, sr);
		
		xmlBody.addXmlChild(xmlSr);
		addSessionId(xmlSr, sr);
		return marshal(xmlEnv);
	}
	
	@Override
	public InputStream marshalSearchRequest(SearchRequest sr)
			throws MarshalException {
		XmlMarshalable xmlEnv = createSoapEnvelope();
		XmlMarshalable xmlBody = attachSoapBody(xmlEnv);
		XmlMarshalable xmlSr = createIfmapNamespaceMarshalable(
				IfmapStrings.SEARCH_REQ_EL_NAME);
	
		xmlBody.addXmlChild(xmlSr);
		addSearchInfo(xmlSr, sr);
		addSessionId(xmlSr, sr);
		return marshal(xmlEnv);
	}

	private void transformSubscribeRequest(XmlMarshalable xmlSr, SubscribeRequest sr)
			throws MarshalException {
		
		if (sr.getSubscribeElements().size() == 0)
			throw new MarshalException("empty subscribe element");
		
		for (SubscribeElement el : sr.getSubscribeElements()) {
			XmlMarshalable xmlEl;
			
			if (el.getName() == null)
				throw new MarshalException("subscription name not set");
			
			if (el instanceof SubscribeUpdate) {
				xmlEl = createNoNamespaceMarshalable(
						IfmapStrings.SUBSCRIBE_UPDATE_EL_NAME);
			} else if (el instanceof SubscribeDelete) {
				xmlEl = createNoNamespaceMarshalable(
						IfmapStrings.SUBSCRIBE_DELETE_EL_NAME);
			} else {
				throw new MarshalException("Unknown SubscribeElement implementation");
			}
			
			xmlEl.addXmlAttribute(IfmapStrings.SUBSCRIBE_SUB_NAME_ATTR, el.getName());
			
			if (el instanceof SearchHolder)
				addSearchInfo(xmlEl, (SearchHolder)el);
			
			xmlSr.addXmlChild(xmlEl);
		}
	}

	private void addSearchInfo(XmlMarshalable xmlEl, SearchHolder el)
			throws MarshalException {
		
		Identifier id = el.getStartIdentifier();
		String matchLinks = el.getMatchLinksFilter();
		String resultFilter = el.getResultFilter();
		Integer maxDepth = el.getMaxDepth();
		Integer maxSize = el.getMaxSize();
		String terminalIdentifiers = el.getTerminalIdentifierTypes();
		
		if (id == null)
			throw new MarshalException("No start Identifier set");
		
		transformAndAddIdentifier(xmlEl, id);
		
		if (matchLinks != null)
			xmlEl.addXmlAttribute(IfmapStrings.SEARCH_MATCH_LINKS_FILTER_ATTR,
					matchLinks);
		
		if (resultFilter != null)
			xmlEl.addXmlAttribute(IfmapStrings.SEARCH_RESULT_FILTER_ATTR,
					resultFilter);
		
		if (terminalIdentifiers != null)
			xmlEl.addXmlAttribute(IfmapStrings.SEARCH_TERM_IDENT_TYPE,
					terminalIdentifiers);
			
		
		if (maxDepth != null)
			xmlEl.addXmlAttribute(IfmapStrings.SEARCH_MAX_DEPTH_ATTR,
					maxDepth.toString());

		if (maxSize != null)
			xmlEl.addXmlAttribute(IfmapStrings.SEARCH_MAX_SIZE_ATTR,
					maxSize.toString());
		
		addXmlNamespaceDeclarations(xmlEl, el);
	}

	private void transformPublishRequest(XmlMarshalable xmlPr, PublishRequest pr)
			throws MarshalException {
		
		if (pr.getPublishElements().size() == 0)
			throw new MarshalException("empty publish element");
		
		for (PublishElement el : pr.getPublishElements()) {
			XmlMarshalable xmlEl;
			
			if (el instanceof PublishUpdate) {
				PublishUpdate pu = (PublishUpdate)el;
				MetadataLifetime lifetime = MetadataLifetime.session;
				xmlEl = createNoNamespaceMarshalable(
						IfmapStrings.PUBLISH_UPDATE_EL_NAME);
				
				if (pu.getLifeTime() != null)
					lifetime = pu.getLifeTime();
				
				xmlEl.addXmlAttribute(IfmapStrings.PUBLISH_UPDATE_ATTR_LIFETIME,
						lifetime.toString());
				
			} else if (el instanceof PublishDelete) {
				PublishDelete pd = (PublishDelete)el;
				xmlEl = createNoNamespaceMarshalable(
						IfmapStrings.PUBLISH_DELETE_EL_NAME);
				if (pd.getFilter() != null)
					xmlEl.addXmlAttribute(IfmapStrings.PUBLISH_DELETE_ATTR_FILTER,
							pd.getFilter());
				
				addXmlNamespaceDeclarations(xmlEl, pd);
				
			} else if (el instanceof PublishNotify) {
				xmlEl = createNoNamespaceMarshalable(
						IfmapStrings.PUBLISH_NOTIFY_EL_NAME);
			} else {
				throw new MarshalException("unknown PublishElement implementation");
			}
			
			if (el instanceof IdentifierHolder)
				transformAndAddIdentifiers(xmlEl, (IdentifierHolder)el);
			else
				throw new MarshalException("not a subclass of IdentifierHoldingRequest?");
			
			if (el instanceof MetadataHolder)
				addMetadataList(xmlEl, (MetadataHolder)el);
			
			xmlPr.addXmlChild(xmlEl);
		}
	}

	
	

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 *                     All the lovely helpers ;)                           *
	 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


	private XmlMarshalable createSoapEnvelope() {
		XmlMarshalable env = new BasicXmlMarshalableImpl(
				IfmapStrings.SOAP_ENV_EL_NAME,
				DomNamespaces.SOAP_NS,
				IfmapStrings.EMPTY_VALUE);
		return env;
	}
	
	private XmlMarshalable attachSoapBody(XmlMarshalable env) {
		
		XmlMarshalable body = new BasicXmlMarshalableImpl(
				IfmapStrings.SOAP_BODY_EL_NAME,
				DomNamespaces.SOAP_NS,
				IfmapStrings.EMPTY_VALUE);
		
		env.addXmlChild(body);
		return body;
	}
	
	private XmlMarshalable createIfmapNamespaceMarshalable(String name) {
		return new BasicXmlMarshalableImpl(
				name,
				DomNamespaces.IFMAP_NS,
				IfmapStrings.EMPTY_VALUE);
	}
	
	private XmlMarshalable createNoNamespaceMarshalable(String name) {
		return new BasicXmlMarshalableImpl(
				name,
				DomNamespaces.NO_NS,
				IfmapStrings.EMPTY_VALUE);
		
	}

	private void addSessionId(XmlMarshalable xml, Request r) {
		if (r.getSessionId() == null)
			throw new NullPointerException("sessionId is null");
		
		xml.addXmlAttribute(IfmapStrings.SESSION_ID_ATTR, r.getSessionId());
	}
	
	private InputStream marshal(XmlMarshalable m) throws MarshalException {
		Document doc = mDocumentbuilder.newDocument();
		
		Element root = createElement(doc, m);
		doc.appendChild(root);
		prepareElementNode(doc, root, m);
		
		return toInputStream(doc);
	}


	private void prepareElementNode(Document doc, Element curEl, XmlMarshalable curM)
			throws MarshalException {
		
		String prefix = curM.getXmlNamespace().getPrefix();
		
		// if length is 0 use null else use the prefix
		prefix = (prefix.length() == 0) ? null : prefix;
		

		// add all namespace declarations as simple attributes
		for (XmlNamespace ns : curM.getXmlNamespaceDeclarations()) {
			String name, value;
			
			name = "xmlns";
			if (ns.getPrefix().length() > 0)
				name += ":" + ns.getPrefix();
			value = ns.getUri();
			curEl.setAttributeNS("http://www.w3.org/2000/xmlns/", name, value);
		}

		for (XmlAttribute attr : curM.getXmlAttributes())
			curEl.setAttributeNS(null, attr.getName(), attr.getValue());

		// add text node if there is one.
		if (curM.getXmlElementValue().length() > 0) {
			Node text = doc.createTextNode(curM.getXmlElementValue());
			curEl.appendChild(text);
		}
		
		// add all XmlMarshalable children
		for (XmlMarshalable child : curM.getXmlChildren()) {
			Element next = createElement(doc, child);
			curEl.appendChild(next);
			prepareElementNode(doc, next, child);
		}
		
		// add all Document children
		for (Document childDoc : curM.getXmlChildrenDoc()) {
			Element clone = (Element) childDoc.getFirstChild();
			if (clone == null)
				throw new MarshalException("Metadata Document without root node");
			
			clone = (Element) clone.cloneNode(true);
			doc.adoptNode(clone);
			curEl.appendChild(clone);
		}
	}
	
	
	private Element createElement(Document doc, XmlMarshalable m) {
		String prefix = m.getXmlNamespace().getPrefix();
		String uri = m.getXmlNamespace().getUri();
		String elementName = m.getXmlElementName();

		elementName = (prefix.length() == 0) ? elementName : prefix + ":" + elementName;
		uri = (prefix.length() == 0) ? null : uri;
		return doc.createElementNS(uri, elementName);
	}

	
	private InputStream toInputStream(Document doc) throws MarshalException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		Result result = new StreamResult(baos);
		byte[] marshalled = null;
		
		doc.setXmlStandalone(true);
		Source source = new DOMSource(doc);

		
		try {
			trans.transform(source, result);
		} catch (TransformerException e) {
			IfmapJLog.error("Oh oh.... [" + e.getMessage() + "]");
			throw new MarshalException(e.getMessage());
		}
		try {
			baos.flush();
		} catch (IOException e) {
			IfmapJLog.error("Oh oh.... [" + e.getMessage() + "]");
			throw new MarshalException(e.getMessage());
		}
	
		marshalled = baos.toByteArray();
		
		return new ByteArrayInputStream(marshalled);
	}
	

	private void addXmlNamespaceDeclarations(XmlMarshalable xmlEl,
			NamespaceDeclarationHolder el) {
		for (Pair<String, String> nsDecl : el.getNamespaceDeclarations())
			xmlEl.addXmlNsDeclaration(new XmlNamespace(nsDecl.first, nsDecl.second));
	}


	
	
	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 *                All the lovely identifier helpers ;)                     *
	 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
	
	
	private void transformAndAddIdentifiers(XmlMarshalable xmlEl,
			IdentifierHolder el) throws MarshalException {
		
		Identifier id1, id2;
		id1 = el.getIdentifier1();
		id2 = el.getIdentifier2();
		
		if (id1 == null && id2 == null)
			throw new MarshalException("need at least one identifier");
		
		if (id1 != null)
			transformAndAddIdentifier(xmlEl, id1);
		
		if (id2 != null)
			transformAndAddIdentifier(xmlEl, id2);
	}

	private void transformAndAddIdentifier(XmlMarshalable xmlEl, Identifier id)
			throws MarshalException {
		
		XmlMarshalable xmlId;
		
		if (id instanceof AccessRequest) {
			xmlId = transformAr((AccessRequest)id);
		} else if (id instanceof Device) {
			xmlId = transformDev((Device)id);
		} else if (id instanceof Identity) {
			xmlId = transformIdentity((Identity)id);
		} else if (id instanceof IpAddress) {
			xmlId = transformIp((IpAddress)id);
		} else if (id instanceof MacAddress) {
			xmlId = transformMac((MacAddress)id);
		} else {
			throw new MarshalException("unknown identifier implementation");
		}
			
		if (id instanceof IdentifierWithAd)
			addAdministrativeDomainInfo(xmlId, (IdentifierWithAd)id);
		
		xmlEl.addXmlChild(xmlId);
	}
	

	private XmlMarshalable transformAr(AccessRequest ar) {
		String name = ar.getName();
		XmlMarshalable ret = createNoNamespaceMarshalable(
				IfmapStrings.ACCESS_REQUEST_EL_NAME);
		
		ret.addXmlAttribute(IfmapStrings.ACCESS_REQUEST_ATTR_NAME, name);
		
		return ret;
	}

	
	private XmlMarshalable transformDev(Device dev) {
		String name = dev.getName();
		XmlMarshalable ret = createNoNamespaceMarshalable(
				IfmapStrings.DEVICE_EL_NAME);
		
		ret.addXmlChild(new BasicXmlMarshalableImpl(
				IfmapStrings.DEVICE_NAME_EL_NAME,
				DomNamespaces.NO_NS,
				name));
		
		return ret;
	}

	private XmlMarshalable transformMac(MacAddress mac) {
		String value = mac.getValue();
		XmlMarshalable ret = createNoNamespaceMarshalable(
				IfmapStrings.MAC_ADDRESS_EL_NAME);
		
		ret.addXmlAttribute(IfmapStrings.MAC_ADDRESS_ATTR_VALUE, value);
		
		return ret;
	}

	private XmlMarshalable transformIp(IpAddress ip) {
		String value = ip.getValue();
		IpAddressType type = IpAddressType.IPv4;
		
		XmlMarshalable ret = createNoNamespaceMarshalable(
				IfmapStrings.IP_ADDRESS_EL_NAME);
		
		type = (ip.getType() != null) ? ip.getType() : type;
		
		ret.addXmlAttribute(IfmapStrings.IP_ADDRESS_ATTR_VALUE, value);
		ret.addXmlAttribute(IfmapStrings.IP_ADDRESS_ATTR_TYPE, type.toString());
		
		return ret;
	}

	private XmlMarshalable transformIdentity(Identity id) throws MarshalException {
		IdentityType type = id.getType();
		String otherTypeDef = id.getOtherTypeDefinition();
		String name = id.getName();
		XmlMarshalable ret = createNoNamespaceMarshalable(
				IfmapStrings.IDENTITY_EL_NAME);
		
		if (type == null)
			throw new MarshalException("No Identity type set");
		
		ret.addXmlAttribute(IfmapStrings.IDENTITY_ATTR_TYPE, type.toString());
		
		ret.addXmlAttribute(IfmapStrings.IDENTITY_ATTR_NAME, name);
		
		if (type == IdentityType.other) {
			if (otherTypeDef == null)
				throw new MarshalException("Identity type=other requires " +
						"other-type-definition");
			ret.addXmlAttribute(IfmapStrings.IDENTITY_ATTR_OTHER_TYPE_DEF, otherTypeDef);
		}
		
		return ret;
	}
		

	private void addAdministrativeDomainInfo(XmlMarshalable xmlId, IdentifierWithAd id) {
		String adminDom = id.getAdministrativeDomain();
		if (adminDom != null && adminDom.length() > 0)
			xmlId.addXmlAttribute(IfmapStrings.IDENTIFIER_ATTR_ADMIN_DOMAIN, adminDom);
	}

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 *                          Lovely Metadata helper ;)                      *
	 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
	
	private void addMetadataList(XmlMarshalable xmlEl, MetadataHolder el)
			throws MarshalException {
		Collection<Document> metadata = el.getMetadata();
		
		if (metadata.size() == 0)
			throw new MarshalException("no metadata in metadata list");
		
		XmlMarshalable xmlMd = new BasicXmlMarshalableImpl(
				IfmapStrings.PUBLISH_METADATA_EL_NAME,
				DomNamespaces.NO_NS,
				IfmapStrings.EMPTY_VALUE);
		
		xmlEl.addXmlChild(xmlMd);
		
		for (Document md : metadata)
			xmlMd.addXmlChildDoc(md);
	}
}
