package de.fhhannover.inform.trust.ifmapj.channel;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.SSLSocketFactory;

import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;

public abstract class CommunicationHandlerFactory {
	
	public static final String HANDLER_PROPERTY = "ifmapj.communication.handler";
	
	/**
	 * Create a {@link CommunicationHandler} instance.
	 * 
	 * The {@link CommunicationHandler} implementation returned by this method
	 * depends on the system property ifmapj.communication.handler. If it is
	 * set to <b>java</b> an instance of {@link JavaCommunicationHandler} is
	 * returned, if set to <b>apache</b> instead, {@link ApacheCoreCommunicationHandler}
	 * is used. If it is not set, either one is returned, preferring
	 * {@link ApacheCoreCommunicationHandler}.
	 * 
	 * @param url The url to connect to.
	 * @param user for basic auth, if user != null -> pass != null
	 * @param pass for basic auth, if pass != null -> user != null
	 * @param sslSocketFactory the {@link SSLSocketFactory} to be used
	 * @return the new {@link CommunicationHandler}
	 * @throws InitializationException 
	 */
	public static CommunicationHandler newHandler(String url, String user,
			String pass, SSLSocketFactory sslSocketFactory, HostnameVerifier verifier)
			throws InitializationException {
		
		String handler = System.getProperty(HANDLER_PROPERTY);
		
		if (handler != null)
			return newHandlerPreference(handler, url, user, pass, sslSocketFactory, verifier);
		else
			return newHandlerAuto(url, user, pass, sslSocketFactory, verifier);
	}
	
	
	/**
	 * Helper to return an instance of the {@link ApacheCoreCommunicationHandler}.
	 * If the {@link ApacheCoreCommunicationHandler} cannot be initialized, return
	 * a {@link JavaCommunicationHandler} instance.
	 * 
	 * @param url
	 * @param user
	 * @param pass
	 * @param sslSocketFactory
	 * @param verifier 
	 * @return
	 * @throws InitializationException
	 */
	private static CommunicationHandler newHandlerAuto(String url, String user,
			String pass, SSLSocketFactory sslSocketFactory, HostnameVerifier verifier) throws InitializationException {
		try {
			return new ApacheCoreCommunicationHandler(url, user, pass,
					sslSocketFactory, verifier);
		} catch (NoClassDefFoundError e) {
			return new JavaCommunicationHandler(url, user, pass,
					sslSocketFactory, verifier);
		}
	}
	
	/**
	 * Helper to return the handler which is indicated by handlerProp
	 * 
	 * @param handlerProp
	 * @param url
	 * @param user
	 * @param pass
	 * @param sslSocketFactory
	 * @param verifier 
	 * @return
	 * @throws InitializationException 
	 */
	private static CommunicationHandler newHandlerPreference(String handlerProp,
			String url, String user, String pass,
			SSLSocketFactory sslSocketFactory, HostnameVerifier verifier) throws InitializationException {
		if (handlerProp == null)
			throw new NullPointerException();
		
		if (handlerProp.equals("java")) {
			return new JavaCommunicationHandler(url, user, pass,
					sslSocketFactory, verifier);
		} else if (handlerProp.equals("apache")) {
			try {
				return new ApacheCoreCommunicationHandler(url, user, pass,
						sslSocketFactory, verifier);
			} catch (NoClassDefFoundError e) {
				throw new InitializationException("Could not initialize ApacheCommunicationHandler");
			}
		} else {
			throw new InitializationException("Invalid " + HANDLER_PROPERTY + " value");
		}
	}
}
